/*
 * Copyright (c) 2006, Adam Dunkels
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */


//------------------------------------------------------------
// Changelog
//------------------------------------------------------------
//  [V:0.1 / 04.09.2013] :
// Author   : Adam Dunkels
// Web      : http://dunkels.com/adam/ubasic/
//
// original Version from Adam Dunkels without changings
//------------------------------------------------------------
//  [V:1.0 / 16.05.2014] :
// Author   : UB
// EMail    : mc-4u(@)t-online.de
// Web      : www.mikrocontroller-4u.de
//
// modified : removed all debug statements
// modified : max len of numbers from 4 to 5 digits
// addded   : new Token "LINENUM"
// added    : new Token "step"
// added    : new Token "rem"
// added    : Tokens not case sensitive
// added    : new Token for System-Function
//             "pause","clrtic","gettic"
// added    : new Tokens for Display-Functions
//             "inverse","cls","lcd","setpx","clrpx","getpx",
//             "line","rect","fill","circle","clrcoll","getcoll"
// added    : new Tokens for GPIO-Functions
//             "in","out"
// added    : new Token for ADC-Function
//             "adc"
//
// BIG CHANGE (to increase the execution speed) :
//
//            Pre-Parse the complete program in 3 Steps
//         
//            1. exchange the ascii-keyword-tokens
//               with an identifier and an array-nummber
//            2. search GOTO/GOSUB/FOR linenumbers and save them
//               in an array
//            3. search for all linenumbers in the array the
//               entry point adress for a fast-jump
//
//------------------------------------------------------------




#include "tokenizer.h"
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>


char *ptr, *nextptr;

#define MAX_NUMLEN 6

struct keyword_token {
  char *keyword;
  int token;
};


static int current_token = TOKENIZER_ERROR;

//------------------------------------------------------------
// added by UB for pre-parser
//------------------------------------------------------------
static int akt_token=TOKENIZER_CR;  
static int parser_mode=0;
int check_preparse(char *program);
int exchange_keywords(char *program);
int search_goto_gosub_for(char *program);  
int search_linenum(char *program);
int check_linenum(int linenum,int mode);  
static int failed, akt_linenum;
struct linenum_buffer {
  int line_nr;
  char *lineptr;
};
#define MAX_LINENUM_BUFFER_DEPTH  100
static struct linenum_buffer linenum_stack[MAX_LINENUM_BUFFER_DEPTH];
static int linenum_stack_ptr;
//------------------------------------------------------------



static const struct keyword_token keywords[] = {
  {"let", TOKENIZER_LET},
  {"print", TOKENIZER_PRINT},
  {"if", TOKENIZER_IF},
  {"then", TOKENIZER_THEN},
  {"else", TOKENIZER_ELSE},
  {"for", TOKENIZER_FOR},
  {"to", TOKENIZER_TO},
  {"next", TOKENIZER_NEXT},
  {"goto", TOKENIZER_GOTO},
  {"gosub", TOKENIZER_GOSUB},
  {"return", TOKENIZER_RETURN},
  {"end", TOKENIZER_END},
  //------------------------------------
  // NEW TOKENS BY UB
  //------------------------------------
  {"step", TOKENIZER_STEP},
  {"rem", TOKENIZER_REM},
  {"pause", TOKENIZER_PAUSE},
  {"clrtic", TOKENIZER_CLRTIC},
  {"gettic", TOKENIZER_GETTIC},
  {"inverse", TOKENIZER_INVERSE},
  {"cls", TOKENIZER_CLS},
  {"lcd", TOKENIZER_PRINT_LCD},
  {"setpx", TOKENIZER_SETPIXEL},
  {"clrpx", TOKENIZER_CLRPIXEL},
  {"getpx", TOKENIZER_GETPIXEL},
  {"line", TOKENIZER_LINE},
  {"rect", TOKENIZER_RECT},
  {"fill", TOKENIZER_FILL},
  {"circle", TOKENIZER_CIRCLE},
  {"clrcoll", TOKENIZER_CLRCOLL},
  {"getcoll", TOKENIZER_GETCOLL},
  {"in", TOKENIZER_IN},
  {"out", TOKENIZER_OUT},
  {"adc", TOKENIZER_ADC},
  //------------------------------------
  {NULL, TOKENIZER_ERROR}
};

/*---------------------------------------------------------------------------*/
static int
singlechar(void)
{
  if(*ptr == '\n') {
    return TOKENIZER_CR;
  } else if(*ptr == ',') {
    return TOKENIZER_COMMA;
  } else if(*ptr == ';') {
    return TOKENIZER_SEMICOLON;
  } else if(*ptr == '+') {
    return TOKENIZER_PLUS;
  } else if(*ptr == '-') {
    return TOKENIZER_MINUS;
  } else if(*ptr == '&') {
    return TOKENIZER_AND;
  } else if(*ptr == '|') {
    return TOKENIZER_OR;
  } else if(*ptr == '*') {
    return TOKENIZER_ASTR;
  } else if(*ptr == '/') {
    return TOKENIZER_SLASH;
  } else if(*ptr == '%') {
    return TOKENIZER_MOD;
  } else if(*ptr == '(') {
    return TOKENIZER_LEFTPAREN;
  } else if(*ptr == ')') {
    return TOKENIZER_RIGHTPAREN;
  } else if(*ptr == '<') {
    return TOKENIZER_LT;
  } else if(*ptr == '>') {
    return TOKENIZER_GT;
  } else if(*ptr == '=') {
    return TOKENIZER_EQ;
  }
  return 0;
}
/*---------------------------------------------------------------------------*/
static int
get_next_token(void)
{
  struct keyword_token const *kt;
  int i,ok;

  if(*ptr == 0) {
    akt_token=TOKENIZER_ENDOFINPUT;
    return akt_token;
  }
	
  if(parser_mode==0) {
    // during pre-parse
    if(*ptr==0x01) {
      // preparse Keyword-Token
      // Basic-Program is already pre-parsed	
      akt_token=TOKENIZER_ENDOFINPUT;
      return akt_token;      
    } else if(isdigit(*ptr)) {
      for(i = 0; i < MAX_NUMLEN; ++i) {
        if(!isdigit(ptr[i])) {
          if(i > 0) {
            nextptr = ptr + i;
            if(akt_token==TOKENIZER_CR) {
              // first number in line = linenum
              akt_token=TOKENIZER_LINENUM;
            }
            else {
              akt_token=TOKENIZER_NUMBER;
            }
            return akt_token;
          } else {
            akt_token=TOKENIZER_ERROR;
            return akt_token;
          }
        }
        if(!isdigit(ptr[i])) {
          akt_token=TOKENIZER_ERROR;
          return akt_token;
        }
      }
      akt_token=TOKENIZER_ERROR;
      return akt_token;
    } else if(singlechar()) {
      nextptr = ptr + 1;
      akt_token=singlechar();
      return akt_token;
    } else if(*ptr == '"') {
      nextptr = ptr;
      do {
        ++nextptr;
      } while(*nextptr != '"');
      ++nextptr;
      akt_token=TOKENIZER_STRING;
      return akt_token;
    } else {
      for(kt = keywords; kt->keyword != NULL; ++kt) {
        if(strncmp(ptr, kt->keyword, strlen(kt->keyword)) == 0) {
          nextptr = ptr + strlen(kt->keyword);
          akt_token=kt->token;
          return akt_token;
        }
        else {
          // check upercase & lowercase
          ok=0;
          for(i=0;i<strlen(kt->keyword);i++) {
            if((*(ptr+i)!=kt->keyword[i]) && (*(ptr+i)!=(kt->keyword[i]-0x20))) {
              ok=1;
              break;
            }
          }
          if(ok==0) {
            nextptr = ptr + strlen(kt->keyword);
            akt_token=kt->token;
            return akt_token;
          }
        }
      }
    }  

    if(*ptr >= 'a' && *ptr <= 'z') {
      nextptr = ptr + 1;
      akt_token=TOKENIZER_VARIABLE;
      return akt_token;
    }

    akt_token=TOKENIZER_ERROR;
    return akt_token; 
  }
  else {
    // after pre-parse
    if(*ptr==0x01) {
      // preparse Keyword-Token
      i=(*(ptr+1)&0x7F);
      nextptr = ptr + strlen(keywords[i].keyword);
      akt_token=keywords[i].token;
      return akt_token;
    } else if(*ptr >= 'a' && *ptr <= 'z') {
      nextptr = ptr + 1;
      akt_token=TOKENIZER_VARIABLE;
      return akt_token;	
    }			
    else if(isdigit(*ptr)) {
      for(i = 0; i < MAX_NUMLEN; ++i) {
        if(!isdigit(ptr[i])) {
          if(i > 0) {
            nextptr = ptr + i;
            if(akt_token==TOKENIZER_CR) {
              // first number in line = linenum
              akt_token=TOKENIZER_LINENUM;
            }
            else {
              akt_token=TOKENIZER_NUMBER;
            }
            return akt_token;
          } else {
            akt_token=TOKENIZER_ERROR;
            return akt_token;
          }
        }
        if(!isdigit(ptr[i])) {
          akt_token=TOKENIZER_ERROR;
          return akt_token;
        }
      }
      akt_token=TOKENIZER_ERROR;
      return akt_token;
    } else if(singlechar()) {
      nextptr = ptr + 1;
      akt_token=singlechar();
      return akt_token;
    } else if(*ptr == '"') {
      nextptr = ptr;
      do {
        ++nextptr;
      } while(*nextptr != '"');
      ++nextptr;
      akt_token=TOKENIZER_STRING;
      return akt_token;
    }		
		
    akt_token=TOKENIZER_ERROR;
    return akt_token;		
  }
}
/*---------------------------------------------------------------------------*/
void
tokenizer_init(char *program)
{
  ptr = program;
  akt_token=TOKENIZER_CR;
  failed =0;
  akt_linenum =0;		
  current_token = get_next_token();
}
/*---------------------------------------------------------------------------*/
int
tokenizer_token(void)
{
  return current_token;
}
/*---------------------------------------------------------------------------*/
void
tokenizer_next(void)
{

  if(tokenizer_finished()) {
    return;
  }

  ptr = nextptr;
  while(*ptr == ' ') {
    ++ptr;
  }
  current_token = get_next_token();
}
/*---------------------------------------------------------------------------*/
int
tokenizer_num(void)
{
  return atoi(ptr);
}
/*---------------------------------------------------------------------------*/
void
tokenizer_string(char *dest, int len)
{
  char *string_end;
  int string_len;

  if(tokenizer_token() != TOKENIZER_STRING) {
    return;
  }
  string_end = strchr(ptr + 1, '"');
  if(string_end == NULL) {
    return;
  }
  string_len = string_end - ptr - 1;
  if(len < string_len) {
    string_len = len;
  }
  memcpy(dest, ptr + 1, string_len);
  dest[string_len] = 0;
}
/*---------------------------------------------------------------------------*/
void
tokenizer_error_print(void)
{

}
/*---------------------------------------------------------------------------*/
int
tokenizer_finished(void)
{
  return *ptr == 0 || current_token == TOKENIZER_ENDOFINPUT;
}
/*---------------------------------------------------------------------------*/
int
tokenizer_variable_num(void)
{
  return *ptr - 'a';
}


/*---------------------------------------------------------------------------*/
// New functions by UB
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
int
tokenizer_pre_parse(char *program)
{
  int status;
	
  tokenizer_init(program);
  // 1.step
  status=exchange_keywords(program);
  if(status!=1) return(1);
  tokenizer_init(program);
  // 2.step
  status=search_goto_gosub_for(program);
  if(status!=1) return(2);
  tokenizer_init(program);
  // 3.step
  status=search_linenum(program);
  if(status!=1) return(3);
  tokenizer_init(program);
	
  return 0;
}


/*---------------------------------------------------------------------------*/
int
exchange_keywords(char *program)
{
  int akt_token=-1,ok=0;
  char *ptr_pos;
  int kpos;	
	
  parser_mode=0;
	
  ptr_pos=program;
	
  ok=0;
  do {
    akt_token=tokenizer_token();
    if(akt_token==TOKENIZER_ERROR) {
      ok=2;
    }
    else if(akt_token==TOKENIZER_ENDOFINPUT) {
      ok=1;
    }
    else if(akt_token==TOKENIZER_LINENUM) {
      if(failed==0) akt_linenum = tokenizer_num();
    }
    else {
      for(kpos=0;kpos<100;kpos++) {
        if(keywords[kpos].token==akt_token) {
          // Keyword-Token
          ptr_pos=ptr;
          *ptr_pos=0x01; // keyword-ID
          ptr_pos++;
          *ptr_pos=(kpos|0x80); // set bit7 to mark as symbol
          break;
        }
        else if(keywords[kpos].token==TOKENIZER_ERROR) {
          break;
        }
      }
    }

    if(ok==0) {
      if(akt_token==TOKENIZER_REM) {
        tokenizer_skipline();
      }
      else {
        tokenizer_next();
      }
    }
  }
  while(ok==0);	
	
  parser_mode=1;
	
  return ok;
}

/*---------------------------------------------------------------------------*/
int
search_goto_gosub_for(char *program)
{
  int akt_token=-1,ok=0,n,merker=0;
	
  linenum_stack_ptr=0;
	
  ok=0;
	merker=0;
  do {
    akt_token=tokenizer_token();
    if(akt_token==TOKENIZER_ERROR) {
      ok=2;
    }
    else if(akt_token==TOKENIZER_ENDOFINPUT) {
      ok=1;
    }
    else if(akt_token==TOKENIZER_LINENUM) {
      n=tokenizer_num();
      if(failed==0) akt_linenum = n;
      if(merker==1) {  // linenum after "FOR" or "GOSUB"
        if(linenum_stack_ptr<MAX_LINENUM_BUFFER_DEPTH) {
          if(check_linenum(n,0)==0) {
            linenum_stack[linenum_stack_ptr].line_nr=n;
            linenum_stack_ptr++;
          }
        } 
        merker=0;
      }
    }
    else if(akt_token==TOKENIZER_GOTO) {
      tokenizer_next();
      akt_token=tokenizer_token();
      if(akt_token==TOKENIZER_NUMBER) { // save linenum from "GOTO"
        if(linenum_stack_ptr<MAX_LINENUM_BUFFER_DEPTH) {
          n=tokenizer_num();
          if(check_linenum(n,0)==0) {
            linenum_stack[linenum_stack_ptr].line_nr=n;
            linenum_stack_ptr++;
          } 
        }
      }
    }	
    else if(akt_token==TOKENIZER_GOSUB) {
      tokenizer_next();
      akt_token=tokenizer_token();
      if(akt_token==TOKENIZER_NUMBER) {  // save linenum from "GOSUB"
        if(linenum_stack_ptr<MAX_LINENUM_BUFFER_DEPTH) {
          n=tokenizer_num();
          if(check_linenum(n,0)==0) {
            linenum_stack[linenum_stack_ptr].line_nr=n;
            linenum_stack_ptr++;
          }
          merker=1; // save next linenum after "GOSUB"
        }
      }			
    }	
    else if(akt_token==TOKENIZER_FOR) {
      merker=1; // save next linenum after "FOR"
    }			 
		 
    if(ok==0) {
      if(akt_token==TOKENIZER_REM) {
        tokenizer_skipline();
      }
      else {
        tokenizer_next();
      }
    }
  }
  while(ok==0);		

  return ok;	
}

/*---------------------------------------------------------------------------*/
int
search_linenum(char *program)
{
  int akt_token=-1,ok=0,n;
  int linecnt=0;
	
  ok=0;
  do {
    akt_token=tokenizer_token();
    if(akt_token==TOKENIZER_ERROR) {
      ok=2;
    }
    else if(akt_token==TOKENIZER_ENDOFINPUT) {
      ok=1;
    }
    else if(akt_token==TOKENIZER_LINENUM) {
      n=tokenizer_num();
      n=check_linenum(n,1);
      if(n!=0) linecnt++;

      tokenizer_skipline();
      tokenizer_next();
    }
    else {
      tokenizer_next();
    }
  }while(ok==0);
  
  if(linecnt!=linenum_stack_ptr) ok=3; // error : not all linenumbers found
	
  return ok;
}

/*---------------------------------------------------------------------------*/
int
check_linenum(int linenum, int mode)
{
  int ret_wert=0;
  int n;
	
  for(n=0;n<linenum_stack_ptr;n++) {
    if(linenum_stack[n].line_nr==linenum) {
      if(mode==1) linenum_stack[n].lineptr=ptr; 
      ret_wert=1;
      break;			
    }
  }
	
  return(ret_wert);
}

/*---------------------------------------------------------------------------*/
int
tokenizer_jumpfast(int linenum)
{
  int ret_wert=0;
  int n;
	
  for(n=0;n<linenum_stack_ptr;n++) {
    if(linenum_stack[n].line_nr==linenum) {
      ptr = linenum_stack[n].lineptr;
      akt_token=TOKENIZER_CR;
      current_token = get_next_token();
      ret_wert=1;
      break;
    }
  }	

  return(ret_wert);
}

/*---------------------------------------------------------------------------*/
void
tokenizer_skipline(void)
{
  ptr = nextptr-1;
  while((*ptr != '\n') && (*ptr != 0x00)) {
    ++ptr;
  }
  current_token = get_next_token();
}
/*---------------------------------------------------------------------------*/
int
tokenizer_linenum(void)
{
  return akt_linenum;
}
/*---------------------------------------------------------------------------*/
